!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ##########################
      MODULE MODI_TRIDIAG_TKE
!     ##########################
INTERFACE
!
       SUBROUTINE TRIDIAG_TKE(KKA,KKU,KKL,PVARM,PA,PTSTEP,PEXPL,PIMPL, &
                                  PRHODJ,PSOURCE,PDIAG,PVARP )
!
INTEGER,                INTENT(IN)   :: KKA     !near ground array index  
INTEGER,                INTENT(IN)   :: KKU     !uppest atmosphere array index
INTEGER,                INTENT(IN)   :: KKL     !vert. levels type 1=MNH -1=ARO
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PVARM       ! variable at t-1  
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PA          ! upper diag. elements
REAL,                   INTENT(IN)  :: PTSTEP      ! Double time step
REAL,                   INTENT(IN)  :: PEXPL,PIMPL ! weights of the temporal scheme
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PRHODJ      ! (dry rho)*J
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSOURCE     ! source term of PVAR    
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PDIAG       ! diagonal term linked to
                                                      ! the implicit dissipation
!
REAL, DIMENSION(:,:,:), INTENT(OUT) :: PVARP       ! variable at t+1       
!
END SUBROUTINE TRIDIAG_TKE
!
END INTERFACE
!
END MODULE MODI_TRIDIAG_TKE
!
!
!
!      ########################################################
       SUBROUTINE TRIDIAG_TKE(KKA,KKU,KKL,PVARM,PA,PTSTEP,PEXPL,PIMPL, &
                                  PRHODJ,PSOURCE,PDIAG,PVARP )
!      ########################################################
!
!
!!****   *TRIDIAG_TKE* - routine to solve a time implicit scheme
!!
!!
!!     PURPOSE
!!     -------
!        The purpose of this routine is to give a field PVARP at t+1, by 
!      solving an implicit tridiagonal system obtained by the 
!      discretization of the vertical turbulent diffusion. It should be noted 
!      that the degree of implicitness can be varied (PIMPL parameter) and the
!      sources of evolution other than the turbulent diffusion can be taken
!      into account through the PSOURCE field. PVARP is localized at a mass 
!      point.
!
!!**   METHOD
!!     ------
!!        First, the Right Hand Side of the implicit equation is computed.
!!     It is build as follows:
!!        ZY = PVARM + PTSTEP*PSOURCE + DIFF_EXPLI
!!     where PVARM is the variable at t-dt, PSOURCE the supplementary sources of
!!     PVAR ( and not PVAR * PRHODJ !!) and  DIFF_EXPLI is the explicit part
!!     of the vertical turbulent diffusion. This operator is spatially 
!!     discretized as the implicit one, thus:
!!        DIFF_EXPLI(k) = - PEXPL / PRHODJ(k) * 
!!                       ( PA(k+1) * (PVARM(k+1) - PVARM(k)  )
!!                        -PA(k)   * (PVARM(k)   - PVARM(k-1)) )
!!     For the first level, only the upper part is considered, the lower one 
!!     is replaced by the turbulent surface flux (taken into account in the 
!!     PSOURCE(ikb) term).
!!        DIFF_EXPLI(ikb) = - PEXPL / PRHODJ(ikb) * 
!!                       ( PA(ikb+1) * (PVARM(ikb+1) - PVARM(ikb))  )
!!     For the last level, only the lower part is considered, the upper one 
!!     is replaced by the turbulent flux which is taken equal to 0 
!!     (taken into account in the PSOURCE(ike) term).
!!
!!        DIFF_EXPLI(ike) = + PEXPL / PRHODJ(ike) * 
!!                       ( PA(ike) * (PVARM(ike) - PVARM(ike-1))  )
!!                      
!!        Then, the classical tridiagonal algorithm is used to invert the 
!!     implicit operator. Its matrix is given by:
!!
!!     ( b(ikb)   c(ikb)      0        0        0         0        0        0  )
!!     (   0      a(ikb+1) b(ikb+1) c(ikb+1)    0  ...    0        0        0  ) 
!!     (   0         0     a(ikb+2) b(ikb+2) c(ikb+2).    0        0        0  ) 
!!      .......................................................................
!!     (   0   ...   0     a(k)     b(k)     c(k)         0   ...  0        0  ) 
!!      .......................................................................
!!     (   0         0        0        0        0 ...a(ike-1) b(ike-1) c(ike-1))
!!     (   0         0        0        0        0 ...     0   a(ike)   b(ike)  )
!!
!!     ikb and ike represent the first and the last inner mass levels of the
!!     model. The coefficients are:
!!         
!!          a(k) = PIMPL * PA(k)/PRHODJ(k)   
!!          b(k) = 1 - PIMPL * PA(k)/PRHODJ(k) - PIMPL * PA(k+1)/PRHODJ(k)
!!          c(k) = PIMPL * PA(k+1)/PRHODJ(k)
!!
!!          for all k /= ikb or ike
!!
!!          b(ikb) = 1 - PIMPL * PA(ikb+1)/PRHODJ(ikb)
!!          c(ikb) = PIMPL * PA(ikb+1)/PRHODJ(ikb)
!!             (discretization of the upper part of the implicit operator)
!!          b(ike) = 1 - PIMPL * PA(ike)/PRHODJ(ike)
!!          a(ike) = PIMPL * PA(ike)/PRHODJ(ike)
!!             (discretization of the lower part of the implicit operator)
!!       Finally, the marginal points are prescribed.
!!
!!       All these computations are purely vertical and vectorizations are 
!!     easely achieved by processing all the verticals in parallel.
!!
!!     EXTERNAL
!!     --------
!!
!!       NONE
!!
!!     IMPLICIT ARGUMENTS
!!     ------------------
!!       MODD_PARAMETERS
!!            JPVEXT_TURB: number of vertical external points
!!
!!     REFERENCE
!!     ---------
!!       Book 1 of Meso-NH documentation (chapter Turbulence)
!!       Press et al: Numerical recipes (1986) Cambridge Univ. Press
!!
!!     AUTHOR
!!     ------
!!       Joan Cuxart       * INM and Meteo-France *   
!! 
!!     MODIFICATIONS
!!     -------------
!!       Original        August 29, 1994
!!       Modification :  January 29, 1995 Algorithm written with two 
!!                                        local variables less
!!      (Cuxart, Stein)  August 21, 1995  Bug correction for PRHODJ
!!      (Stein)         November 16, 1995 new version
!!      (Stein)         February 28, 1995 no inversion in the explicit case
!! ---------------------------------------------------------------------
!
!*       0. DECLARATIONS
!
USE MODD_PARAMETERS
USE MODE_MPPDB
!
#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK , ONLY : MNH_ALLOCATE_ZT3D , MNH_REL_ZT3D , MNH_ALLOCATE_ZT2D, &
      MNH_CHECK_IN_ZT3D,MNH_CHECK_OUT_ZT3D
#endif
!
IMPLICIT NONE
!
!
!*       0.1 declarations of arguments
!
INTEGER,              INTENT(IN)   :: KKA     !near ground array index  
INTEGER,              INTENT(IN)   :: KKU     !uppest atmosphere array index
INTEGER,              INTENT(IN)   :: KKL     !vert. levels type 1=MNH -1=ARO
REAL, DIMENSION(:,:,:),    INTENT(IN)  :: PVARM       ! variable at t-1  
REAL, DIMENSION(:,:,:),    INTENT(IN)  :: PA          ! upper diag. elements
REAL,                      INTENT(IN)  :: PTSTEP      ! Double time step
REAL,                      INTENT(IN)  :: PEXPL,PIMPL ! weights of the temporal scheme
REAL, DIMENSION(:,:,:),    INTENT(IN)  :: PRHODJ      ! (dry rho)*J
REAL, DIMENSION(:,:,:),    INTENT(IN)  :: PSOURCE     ! source term of PVAR    
REAL, DIMENSION(:,:,:),    INTENT(IN)  :: PDIAG       ! diagonal term linked to
                                                      ! the implicit dissipation
!
REAL, DIMENSION(:,:,:),    INTENT(OUT) :: PVARP       ! variable at t+1        
!
!*       0.2 declarations of local variables
!
INTEGER             :: JI,JJ,JK      ! loop counters
INTEGER             :: IKB,IKE       ! inner vertical limits
INTEGER             :: IKT          ! array size in k direction
INTEGER             :: IKTB,IKTE    ! start, end of k loops in physical domain
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZY ,ZGAM ! RHS of the equation, 3D work array
REAL, DIMENSION(:,:),   pointer , contiguous :: ZBET     ! 2D work array
#ifdef MNH_OPENACC
INTEGER :: IZY ,IZGAM, IZBET
#endif
!
INTEGER  :: JIU,JJU,JKU
! ---------------------------------------------------------------------------

!$acc data present( PVARM, PA, PRHODJ, PSOURCE, PDIAG, PVARP )

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( pvarm,   "Tridiag_tke beg:pvarm"   )
  call Mppdb_check( pa,      "Tridiag_tke beg:pa"      )
  call Mppdb_check( prhodj,  "Tridiag_tke beg:prhodj"  )
  call Mppdb_check( psource, "Tridiag_tke beg:psource" )
  call Mppdb_check( pdiag,   "Tridiag_tke beg:pdiag"   )
end if

JIU =  size( pvarm, 1 )
JJU =  size( pvarm, 2 )
JKU =  size( pvarm, 3 )

#ifndef MNH_OPENACC
allocate( zy  (JIU,JJU,JKU ) )
allocate( zgam(JIU,JJU,JKU ) )
allocate( zbet(JIU,JJU ) )
#else
CALL MNH_CHECK_IN_ZT3D("TRIDIAG_TKE")
izy   = MNH_ALLOCATE_ZT3D( zy  ,JIU,JJU,JKU )
izgam = MNH_ALLOCATE_ZT3D( zgam,JIU,JJU,JKU )
izbet = MNH_ALLOCATE_ZT2D( zbet,JIU,JJU )
#endif

!$acc data present( ZY, ZGAM, ZBET )

!
!*      1.  COMPUTE THE RIGHT HAND SIDE
!           ---------------------------
!
!$acc kernels
IKT=SIZE(PVARM,3)
IKTB=1+JPVEXT_TURB
IKTE=IKT-JPVEXT_TURB
IKB=KKA+JPVEXT_TURB*KKL
IKE=KKU-JPVEXT_TURB*KKL

!
!
!$acc loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
   ZY(JI,JJ,IKB) = PVARM(JI,JJ,IKB)  + PTSTEP*PSOURCE(JI,JJ,IKB) -   &
        PEXPL / PRHODJ(JI,JJ,IKB) * PA(JI,JJ,IKB+KKL) * (PVARM(JI,JJ,IKB+KKL) - PVARM(JI,JJ,IKB))
END DO !CONCURRENT
!
DO JK=IKTB+1,IKTE-1
   !$acc loop independent collapse(2)
   DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
      ZY(JI,JJ,JK)= PVARM(JI,JJ,JK)  + PTSTEP*PSOURCE(JI,JJ,JK) -               &
           PEXPL / PRHODJ(JI,JJ,JK) *                                           &
           ( PVARM(JI,JJ,JK-KKL)*PA(JI,JJ,JK)                &
           -PVARM(JI,JJ,JK)*(PA(JI,JJ,JK)+PA(JI,JJ,JK+KKL))   &
           +PVARM(JI,JJ,JK+KKL)*PA(JI,JJ,JK+KKL)              &
           )
   END DO !CONCURRENT
END DO
!
!$acc loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
   ZY(JI,JJ,IKE)= PVARM(JI,JJ,IKE) + PTSTEP*PSOURCE(JI,JJ,IKE) +               &
        PEXPL / PRHODJ(JI,JJ,IKE) * PA(JI,JJ,IKE) * (PVARM(JI,JJ,IKE)-PVARM(JI,JJ,IKE-KKL))
END DO !CONCURRENT
!
!
!*       2.  INVERSION OF THE TRIDIAGONAL SYSTEM
!            -----------------------------------
!
IF ( PIMPL > 1.E-10 ) THEN
!
  !
  !  going up
   !
  !$acc loop independent collapse(2)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU) 
     ZBET(JI,JJ) = 1. + PIMPL * (PDIAG(JI,JJ,IKB)-PA(JI,JJ,IKB+KKL) / PRHODJ(JI,JJ,IKB))
     ! bet = b(ikb)
     PVARP(JI,JJ,IKB) = ZY(JI,JJ,IKB) / ZBET(JI,JJ)
  END DO !CONCURRENT
  !
  !$acc loop  seq
  DO JK = IKB+KKL,IKE-KKL,KKL
   !$acc loop gang, vector collapse(2) independent
    DO JJ=1,JJU
      DO JI=1,JIU
        ZGAM(JI,JJ,JK) = PIMPL * PA(JI,JJ,JK) / PRHODJ(JI,JJ,JK-KKL) / ZBET(JI,JJ)  
                                                    ! gam(k) = c(k-1) / bet
        ZBET(JI,JJ)    = 1. + PIMPL * ( PDIAG(JI,JJ,JK) -                 &
                                 (  PA(JI,JJ,JK) * (1. + ZGAM(JI,JJ,JK))  &
                                  + PA(JI,JJ,JK+KKL)                      &
                                 ) / PRHODJ(JI,JJ,JK)                     &
                                )                   ! bet = b(k) - a(k)* gam(k)  
        PVARP(JI,JJ,JK)= ( ZY(JI,JJ,JK) - PIMPL * PA(JI,JJ,JK) / PRHODJ(JI,JJ,JK) &
                    * PVARP(JI,JJ,JK-KKL)                                         &
                   ) / ZBET(JI,JJ)
                                        ! res(k) = (y(k) -a(k)*res(k-1))/ bet 
      END DO
    END DO
  END DO 
  ! special treatment for the last level
  !$acc loop independent collapse(2)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
     ZGAM(JI,JJ,IKE) = PIMPL * PA(JI,JJ,IKE) / PRHODJ(JI,JJ,IKE-KKL) / ZBET(JI,JJ) 
     ! gam(k) = c(k-1) / bet
     ZBET(JI,JJ)    = 1. + PIMPL * ( PDIAG(JI,JJ,IKE) -                   &
          (  PA(JI,JJ,IKE) * (1. + ZGAM(JI,JJ,IKE)) ) / PRHODJ(JI,JJ,IKE) &
          )  
     ! bet = b(k) - a(k)* gam(k)  
     PVARP(JI,JJ,IKE)= ( ZY(JI,JJ,IKE) - PIMPL * PA(JI,JJ,IKE) / PRHODJ(JI,JJ,IKE) &
          * PVARP(JI,JJ,IKE-KKL)                      &
          ) / ZBET(JI,JJ)
     ! res(k) = (y(k) -a(k)*res(k-1))/ bet
  END DO !CONCURRENT
  !
  !  going down
  !
  !$acc loop seq
  DO JK = IKE-KKL,IKB,-1*KKL
     !$acc loop gang, vector collapse(2)
     DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
        PVARP(JI,JJ,JK) = PVARP(JI,JJ,JK) - ZGAM(JI,JJ,JK+KKL) * PVARP(JI,JJ,JK+KKL)
     END DO !CONCURRENT   
  END DO
!
ELSE
  !
  !$acc loop independent collapse(2)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU) 
     PVARP(JI,JJ,IKTB:IKTE) = ZY(JI,JJ,IKTB:IKTE)
  END DO !CONCURRENT   
!
END IF 
!
!
!*       3.  FILL THE UPPER AND LOWER EXTERNAL VALUES
!            ----------------------------------------
!
!$acc loop independent collapse(2)
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU)
   PVARP(JI,JJ,KKA)=PVARP(JI,JJ,IKB)
   PVARP(JI,JJ,KKU)=PVARP(JI,JJ,IKE)
END DO !CONCURRENT
!$acc end kernels

if ( mppdb_initialized ) then
  !Check all out arrays
  call Mppdb_check( pvarp, "Tridiag_tke end:pvarp" )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate( ZY, ZGAM, ZBET )
#else
CALL MNH_REL_ZT3D( IZY, IZGAM, IZBET )
CALL MNH_CHECK_OUT_ZT3D("TRIDIAG_TKE")
#endif

!$acc end data

!-------------------------------------------------------------------------------
!
END SUBROUTINE TRIDIAG_TKE
